@@ -7,7 +7,7 @@ use parent qw{ Exporter };
 
 use vars qw{ @EXPORT_OK };
 
-@EXPORT_OK = qw{ process_file };
+@EXPORT_OK = qw{ process_file post pre process filter next_line };
 
 use Carp;
 use Carp::Always;
@@ -23,74 +23,124 @@ Readonly my $FALSE   => 0;
 Readonly my $EMPTY   => q{};
 Readonly my $NL      => "\n";
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 
 our %DEFAULT_PROCESSORS = (
-  pre => sub {
-    my ( $file, $args ) = @_;
-
-    my $fh;
-
-    if ( openhandle $file ) {
-      $fh = $file;
-
-      $args->{file} = ref $fh;    # GLOB
-    }
-    else {
-      open $fh, '<', $file
-        or croak 'could not open ' . $file . $NL;
-
-      $args->{'file'} = $file;
-    } ## end else [ if ( openhandle($file))]
-
-    $args->{'raw_count'}  = 0;
-    $args->{'skipped'}    = 0;
-    $args->{'start_time'} = time;
-
-    my $lines = $args->{merge_lines} ? IO::Scalar->new : [];
-
-    return ( $fh, $lines );
-  },
-  next_line => sub {
-    my ( $fh, $all_lines, $args ) = @_;
-    my $cur_line;
-
-    if ( openhandle $fh ) {
-      if ( !eof $fh ) {
-        defined( $cur_line = readline $fh )
-          or croak "readline failed: $OS_ERROR\n";
-      }
-    } ## end if ( !eof $fh )
-
-    return $cur_line;
-  },
-  process => sub {
-    my ( $fh, $all_lines, $args, $current_line ) = @_;
-
-    return $current_line;
-  },
-  post => sub {
-    my ( $fh, $all_lines, $args ) = @_;
-
-    $args->{end_time} = time;
-
-    my $retval;
-
-    if ( $args->{merge_lines} ) {
-      $retval = ${ $all_lines->sref };
-    }
-    else {
-      $retval = $all_lines;
-    }
-
-    close $fh
-      or croak 'could not close' . $args->{file} . $NL;
-
-    return $retval, %$args;
-  },
+  pre       => \&_pre,
+  next_line => \&_next_line,
+  filter    => \&_filter,
+  process   => \&_process,
+  post      => \&_post,
 );
 
-caller() or __PACKAGE__->main();
+caller or __PACKAGE__->main();
+
+sub _pre {
+  my ( $file, $args ) = @_;
+
+  my $fh;
+
+  if ( openhandle $file ) {
+    $fh = $file;
+
+    $args->{file} = ref $fh;    # GLOB
+  }
+  else {
+    open $fh, '<', $file        ## no critic [InputOutput::RequireBriefOpen]
+      or croak 'could not open ' . $file . $NL;
+
+    $args->{'file'} = $file;
+  } ## end else [ if ( openhandle $file )]
+
+  $args->{'raw_count'}  = 0;
+  $args->{'skipped'}    = 0;
+  $args->{'start_time'} = time;
+
+  my $lines = $args->{merge_lines} ? IO::Scalar->new : [];
+
+  return ( $fh, $lines );
+} ## end sub _pre
+
+sub _next_line {
+  my ( $fh, $all_lines, $args ) = @_;
+  my $current_line;
+
+  if ( openhandle $fh ) {
+    if ( !eof $fh ) {
+      defined( $current_line = readline $fh )
+        or croak "readline failed: $OS_ERROR\n";
+    }
+  } ## end if ( openhandle $fh )
+
+  return $current_line;
+} ## end sub _next_line
+
+sub _filter {
+  my ( $fh, $all_lines, $args, $current_line ) = @_;
+
+  if ( $args->{'chomp'} ) {
+    chomp $current_line;
+  }
+
+  if ( $args->{'trim'} && $args->{'trim'} =~ /(front|both)/xsm ) {
+    $current_line =~ s/^\s+//xsm;
+  }
+
+  if ( $args->{'trim'} && $args->{'trim'} =~ /(both|back)/xsm ) {
+    $current_line =~ s/\s+$//xsm;
+  }
+
+  # skip?
+  my $skip = $FALSE;
+
+  if ( $args->{'skip_blank_lines'} || $args->{'skip_comments'} ) {
+
+    if ( $args->{'skip_blank_lines'} && "$current_line" eq $EMPTY ) {
+      $skip = $TRUE;
+    }
+
+    # if we're not chomping, then consider new line a blank line?
+    if ( !$args->{chomp} && "$current_line" eq $NL ) {
+      $skip = $TRUE;
+    }
+
+    if ( $args->{'skip_comments'} && $current_line =~ /^\#/xsm ) {
+      $skip = $TRUE;
+    }
+  } ## end if ( $args->{'skip_blank_lines'...})
+
+  $args->{skipped} = $args->{skipped} + $skip ? 1 : 0;
+
+  return $skip ? undef : $current_line;
+} ## end sub _filter
+
+sub _process {
+  my ( $fh, $all_lines, $args, $current_line ) = @_;
+
+  return $current_line;
+}
+
+sub _post {
+  my ( $fh, $all_lines, $args ) = @_;
+
+  $args->{end_time} = time;
+
+  my $retval;
+
+  if ( $args->{merge_lines} ) {
+    $retval = ${ $all_lines->sref };
+  }
+  else {
+    $retval = $all_lines;
+  }
+
+  if ( !$args->{'keep_open'} ) {
+    close $fh
+      or croak 'could not close' . $args->{file} . $NL;
+  }
+
+  return $retval, %{$args};
+} ## end sub _post
 
 sub process_file {
   my ( $file, %args ) = @_;
@@ -99,9 +149,10 @@ sub process_file {
 
   $args{'file'} = $file || $EMPTY;
 
-  my %processors = map { $_, $args{$_} } qw{ pre next_line process post };
+  my %processors
+    = map { ( $_, $args{$_} ) } qw{ pre filter next_line process post };
 
-  foreach (qw{ pre next_line process post}) {
+  foreach (qw{ pre filter next_line process post}) {
     if ( !$processors{$_} ) {
       $processors{$_} = $DEFAULT_PROCESSORS{$_};
     }
@@ -115,80 +166,47 @@ sub process_file {
     croak "invalid pre processor return: wanted file handle, array ref\n";
   }
 
-  while (1) {
-    my $cur_line = $processors{next_line}->( $fh, $all_lines, \%args );
-    last if !defined $cur_line;
+  LINE: while (1) {
+    my $current_line = $processors{'next_line'}->( $fh, $all_lines, \%args );
+    last LINE if !defined $current_line;
 
     $args{'raw_count'}++;
 
-    if ( $args{'chomp'} ) {
-      chomp $cur_line;
-    }
-
-    if ( $args{'trim'} && $args{'trim'} =~ /(front|both)/ ) {
-      $cur_line =~ s/^\s+//;
-    }
-
-    if ( $args{'trim'} && $args{'trim'} =~ /(both|back)/ ) {
-      $cur_line =~ s/\s+$//;
-    }
-
-    # skip?
-    my $skip = $FALSE;
-
-    if ( $args{'skip_blank_lines'} || $args{'skip_comments'} ) {
-
-      if ( $args{'skip_blank_lines'} && "$cur_line" eq $EMPTY ) {
-        $skip = $TRUE;
-      }
-
-      # if we're not chomping, then consider new line a blank line?
-      if ( !$args{chomp} && "$cur_line" eq $NL ) {
-        $skip = $TRUE;
-      }
-
-      if ( $args{'skip_comments'} && $cur_line =~ /^#/xsm ) {
-        $skip = $TRUE;
-      }
-    } ## end if ( $args{'skip_blank_lines'...})
-
-    $args{skipped} = $args{skipped} + $skip ? 1 : 0;
-    next if $skip;
-
-    $cur_line
-      = $processors{'process'}->( $fh, $all_lines, \%args, $cur_line );
-
-    # return undef to halt processing (or die)
-    last if !defined $cur_line;
+    foreach my $p ( @processors{qw{ filter process }} ) {
+      $current_line
+        = eval { return $p->( $fh, $all_lines, \%args, $current_line ); };
+      last LINE if $EVAL_ERROR;
+      next LINE if !defined $current_line;
+    } ## end foreach my $p ( @processors...)
 
     if ( $args{merge_lines} ) {
-      $all_lines->print($cur_line);
+      $all_lines->print($current_line);
     }
     else {
-      push @{$all_lines}, $cur_line;
+      push @{$all_lines}, $current_line;
     }
-  } ## end while (1)
+  } ## end LINE: while (1)
 
   return $processors{'post'}->( $fh, $all_lines, \%args );
 } ## end sub process_file
 
-sub post {
-  my ($fh, $all_lines, $args) = @_;
-  return($_[2]->{default_processors}->{post}->(@_));
+sub post {    ## no critic [Subroutines::RequireArgUnpacking]
+  my ( $fh, $all_lines, $args ) = @_;
+  return ( $_[2]->{default_processors}->{post}->(@_) );
 }
 
-sub pre {
-  my ($fh, $args) = @_;
-  return ($args->{default_processors}->{pre}->(@_));
+sub pre {    ## no critic [Subroutines::RequireArgUnpacking]
+  my ( $fh, $args ) = @_;
+  return ( $args->{default_processors}->{pre}->(@_) );
 }
 
-sub process {
-  my ($fh, $all_lines, $args) = @_;
+sub process {    ## no critic [Subroutines::RequireArgUnpacking]
+  my ( $fh, $all_lines, $args ) = @_;
   return $_[2]->{default_processors}->{process}->(@_);
 }
 
-sub next_line {
-  my ($fh, $all_lines, $args) = @_;
+sub next_line {    ## no critic [Subroutines::RequireArgUnpacking]
+  my ( $fh, $all_lines, $args ) = @_;
   return $_[2]->{default_processors}->{next_line}->(@_);
 }
 
@@ -197,16 +215,16 @@ sub main {
   require Data::Dumper;
   require JSON::PP;
   require Text::CSV_XS;
-  
+
   JSON::PP->import('decode_json');
-  
+
   Data::Dumper->import('Dumper');
-  
+
   # +------------------+
-  # | READ A TEXT FILE | 
+  # | READ A TEXT FILE |
   # +------------------+
 
-  my $buffer =<<'END_OF_TEXT';
+  my $buffer = <<'END_OF_TEXT';
 line 1
  line 2   
  
@@ -215,7 +233,7 @@ line 4
 line 5
 END_OF_TEXT
 
-  my $fh = IO::Scalar->new(\$buffer);
+  my $fh = IO::Scalar->new( \$buffer );
 
   print Dumper(
     process_file(
@@ -226,24 +244,23 @@ END_OF_TEXT
     )
   );
 
-  $fh = IO::Scalar->new(\$buffer);
+  $fh = IO::Scalar->new( \$buffer );
   print Dumper(
     process_file(
       $fh,
-      post  => sub {
+      post => sub {
         my @retval = post(@_);
-        $retval[0] = join '', @{ $_[1] };
+        $retval[0] = join $EMPTY, @{ $_[1] };
         return @retval;
       }
     )
   );
 
-
   # +------------------+
-  # | READ A JSON FILE | 
+  # | READ A JSON FILE |
   # +------------------+
 
-  my $json_text =<<'END_OF_TEXT';
+  my $json_text = <<'END_OF_TEXT';
 {
   "foo" : "bar",
   "baz" : "buz"
@@ -251,30 +268,26 @@ END_OF_TEXT
 
 END_OF_TEXT
 
-  $fh = IO::Scalar->new(\$json_text);
-  
+  $fh = IO::Scalar->new( \$json_text );
+
   print Dumper(
     process_file(
       $fh,
       chomp => 1,
       post  => sub {
         post(@_);
-        return decode_json( join '', @{ $_[1] } );
+        return decode_json( join $EMPTY, @{ $_[1] } );
       }
     )
   );
 
-  $fh = IO::Scalar->new(\$json_text);
-  
-  print Dumper(
-    decode_json(
-      process_file( $fh, merge_lines => 1, chomp => 1)
-    )
-  );
+  $fh = IO::Scalar->new( \$json_text );
 
+  print Dumper(
+    decode_json( process_file( $fh, merge_lines => 1, chomp => 1 ) ) );
 
   # +-----------------+
-  # | READ A CSV FILE | 
+  # | READ A CSV FILE |
   # +-----------------+
 
   my $csv_text = <<'END_OF_TEXT';
@@ -288,49 +301,51 @@ END_OF_TEXT
 
   my ($csv_lines) = process_file(
     $fh,
-    csv   => $csv,
-    chomp => 1,
+    csv         => $csv,
+    chomp       => 1,
     has_headers => 1,
-    pre   => sub {
-      my ( $fh, $args ) = @_;
+    pre         => sub {
+      my ( $csv_fh, $args ) = @_;
 
       if ( $args->{'has_headers'} ) {
-        my @column_names = $args->{csv}->getline($fh);
+        my @column_names = $args->{csv}->getline($csv_fh);
         $args->{csv}->column_names(@column_names);
       }
 
-      return (pre($fh, $args));
+      return ( pre( $fh, $args ) );
     },
     next_line => sub {
-      my ( $fh, $all_lines, $args ) = @_;
-      my $ref = $args->{csv}->getline_hr($fh);
+      my ( $csv_fh, $all_lines, $args ) = @_;
+      my $ref = $args->{csv}->getline_hr($csv_fh);
       return $ref;
-      }
+    }
   );
 
   print Dumper($csv_lines);
-                               
+
   exit 0;
-}
+} ## end sub main
 
 1;
 
+__END__
+
 =pod
 
 =head1 NAME
 
-C<File::Process>
+File::Process - process text files with customer handlers
 
 =head1 SYNOPSIS
 
- my ($lines, $info) = $process_file($file, process => sub { 
+ my ($lines, $info) = process_file($file, process => sub { 
      my ($fh, $lines, $args, $line) = @_;
      return uc($line);
     });
 
 =head1 DESCRIPTION
 
-Many scripts need to process on or more text files. The boiler-plate
+Many scripts need to process one or more text files. The boiler-plate
 usually looks something like:
 
  open my $fh, '<', $file
@@ -344,13 +359,13 @@ usually looks something like:
     croak "blah blah blah...\n";
 
 The I<do something...> part often involves other common operations like
-removing new lines, skipping blank lines, etc. It gets worse when you
+removing new lines, skipping blank lines, etc. It gets tedious when you
 have to write the same template for processing different files in a
 script. 
 
-This class that provides a simple harness for processing text files,
-taking the drudgery out of writing a simple text processor. It is most
-effect when used on relatively small files.
+This class provides a simple harness for processing files, taking
+the drudgery out of writing a simple text processor. It is most effect
+when used on relatively small files.
 
 In it's most basic form the class will return all of the lines in a
 text file. The class exports 1 method (C<process_file>) which uses
@@ -369,7 +384,7 @@ options during the processing of the file.
 The method returns a list containing a reference to an array that
 contains each line of the file followed by the list of elements in the
 hash that was originally passed to it (along with any other data your
-custom method have inserted into it).
+custom method has inserted into it).
 
  my ($lines, %options) = process_file("foo.txt", chomp => 1);
 
@@ -382,15 +397,18 @@ Path to the file to be processed or a handle to an open file.
 =item options
 
 A B<list> of options. You can send whatever options your custom
-processor supports. As each line is read C<process_file> supports
-these options:
+processor supports. Before the default or your custom C<process>
+subroutine is called, the C<filter> subroutine is called. This is
+where you might massage the input in some way.  The default C<filter>
+subroutine supports various options to perform routine tasks. Options
+are described below.
 
 =over 10
 
 =item skip_blank_lines
 
-Skip I<blank lines>. A blank line is considered a line with no data or
-if I<chomp> mode is B<not> enabled, a line with only a new line character.
+Skip I<blank lines>. A blank line is considered a line with only a new
+line character.
 
 =item skip_comments
 
@@ -401,7 +419,8 @@ Set C<skip_lines> to a true value to skip lines that beging with '#'.
 Merges lines together rather that creating an array of
 lines. Typically used with the C<chomp> option. When C<merge_lines> is
 set to a true value, C<IO::Scalar> is used to efficiently create a
-single scalar from all of the lines in the file.
+single scalar from all of the lines in the file.  The first element of
+the return list then a scalar instead of an array reference.
 
 =item chomp
 
@@ -409,7 +428,7 @@ Set C<chomp> to a true value to remove a trailing new line.
 
 =item trim
 
-Set C<trim> to one of I<front>, I<back>, I<both> true value to remove
+Set C<trim> to one of I<front>, I<back>, I<both> to remove
 whitespace from the front, back or both front and back of a line. Note
 that this operation is performed I<before> your custom processor is
 called and may result in the line being skipped if the
@@ -417,12 +436,34 @@ C<skip_blank_lines> option is set to a true value.
 
 =back
 
-C<process_file> will execute a pre processing subroutine (I<pre>), a
-processor (I<process>) which is passed the next line and returns the
-processed line, and a post processing routine (I<post>).  The
-default processors are described below.
+=head2 Custom Processors
 
-=head2 pre(file, options)
+C<process_file> will execute a set of subroutines for each line of the
+file. You can replace any of these subroutines to inject your own
+custom behaviors. They are executed in this order:
+
+=over 5
+
+=item 1. C<pre>
+
+=item 2. C<next_line>
+
+=item 3. C<filter>
+
+=item 4. C<process>
+
+=item 5. C<post>
+
+=back
+
+The default processors are described below.
+
+=head3 pre(file, options)
+
+The default C<pre> processor opens the file and returns a file handle,
+a reference to a array that will be used to store the lines. The
+return should be a tuple that contains the file handle and a reference
+to an array that will be used to store each processed line of the file.
 
 =over 5
 
@@ -441,27 +482,43 @@ potentially useful statistics.
 
 =back
 
-=head2 next_line(fh, lines, options)
+=head3 next_line(fh, lines, options)
 
 The C<next_line> method is passed the file handle, the buffer of
 accumulated lines, and a reference to a hash of options passed to
 C<process_file>. It is expected to return the I<next line> of the
-file. Your custom processor however can return anything it likes. That
-object will be sent to the C<process> subroutine for possible further
-processing.
+file, however your custom processor however can return anything it
+likes. That object returned will be sent to the C<process> subroutine
+for possible further processing.
 
 Returning C<undef> will halt further processing.
 
-=head2 process(fh, lines, options, current_line)
+=head3 filter(fh, lines, options, current_line)
+
+The default C<filter> method will perform various tasks (chomp, trim,
+skip) controlled by the options described above.
+
+If the C<chomp> option is set to true when you called C<process_file>,
+the line will be chomped.  You can also set the C<skip_blank_lines>
+or C<skip_comments> to skip blank lines or skip lines that begin
+with '# '.
+
+Filters should return the line or C<undef> to skip the current line.
+If you really want to add C<undef> to your buffer, do so in your
+filter:
+
+ push @{$lines}, undef;
+
+If you want to halt processing here, C<die> in your filter. Any
+exception will halt further processing.
+
+=head3 process(fh, lines, options, current_line)
 
 The C<process> method is passed the file handle, the buffer of
 accumulated lines, a reference to a hash of options passed to
 C<process_file> and the next line of the the text file.  The default
-processor simply returns the C<current_line> value.  If the C<chomp>
-option is set to true when you called C<process_file>, the line will
-be chomped.  You can also set the C<skip_blank_lines> or
-C<skip_comments> to skip blank lines or skip lines that begin with
-'# '. 
+processor simply returns the C<current_line> value.
+
 
 =over 5
 
@@ -484,7 +541,7 @@ The next line of data from the file.
 
 =back
 
-=head2 post(fh, lines, options)
+=head3 post(fh, lines, options)
 
 The C<post> method is passed the same three arguments as passed to
 C<process>.  The default C<post> method closes the file and records
